from shexer.utils.factories.triple_yielders_factory import get_triple_yielder
from shexer.core.class_profiler import ClassProfiler


def get_class_profiler(target_classes_dict, source_file, list_of_source_files, input_format,
                       instantiation_property_str,
                       namespaces_to_ignore=None,
                       infer_numeric_types_for_untyped_literals=False,
                       raw_graph=None,
                       namespaces_dict=None,
                       url_input=None,
                       list_of_url_input=None,
                       rdflib_graph=None,
                       shape_map_file=None,
                       shape_map_raw=None,
                       track_classes_for_entities_at_last_depth_level=True,
                       depth_for_building_subgraph=1,
                       url_endpoint=None,
                       strict_syntax_with_corners=False,
                       target_classes=None,
                       file_target_classes=None,
                       built_remote_graph=None,
                       built_shape_map=None,
                       remove_empty_shapes=True):
    yielder = get_triple_yielder(source_file=source_file,
                                 list_of_source_files=list_of_source_files,
                                 input_format=input_format,
                                 namespaces_to_ignore=namespaces_to_ignore,
                                 raw_graph=raw_graph,
                                 allow_untyped_numbers=infer_numeric_types_for_untyped_literals,
                                 namespaces_dict=namespaces_dict,
                                 url_input=url_input,
                                 list_of_url_input=list_of_url_input,
                                 rdflib_graph=rdflib_graph,
                                 shape_map_file=shape_map_file,
                                 shape_map_raw=shape_map_raw,
                                 track_classes_for_entities_at_last_depth_level=track_classes_for_entities_at_last_depth_level,
                                 depth_for_building_subgraph=depth_for_building_subgraph,
                                 url_endpoint=url_endpoint,
                                 instantiation_property=instantiation_property_str,
                                 strict_syntax_with_corners=strict_syntax_with_corners,
                                 target_classes=target_classes,
                                 file_target_classes=file_target_classes,
                                 built_remote_graph=built_remote_graph,
                                 built_shape_map=built_shape_map)

    return ClassProfiler(triples_yielder=yielder,
                         target_classes_dict=target_classes_dict,
                         instantiation_property_str=instantiation_property_str,
                         original_target_classes=target_classes,
                         original_shape_map=built_shape_map,
                         remove_empty_shapes=remove_empty_shapes)



from shexer.io.graph.yielder.rdflib_triple_yielder import RdflibParserTripleYielder, RdflibTripleYielder
from rdflib import Graph
from io import StringIO

graph = """
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix dc: <http://purl.org/dc/elements/1.1/> .
@prefix ex: <http://example.org/stuff/1.0/> .

<http://www.w3.org/TR/rdf-syntax-grammar>
  dc:title "RDF/XML Syntax Specification (Revised)" ;
  ex:editor [
    ex:fullname "Dave Beckett";
    ex:homePage <http://purl.org/net/dajobe/>
  ] ."""

yielder = RdflibParserTripleYielder(raw_graph=graph)
for a_triple in yielder.yield_triples():
    print(a_triple)

print("-------------------")

obj_graph = Graph()
obj_graph.parse(StringIO(graph), format="ttl")

yielder = RdflibTripleYielder(rdflib_graph=obj_graph)
for a_triple in yielder.yield_triples():
    print(a_triple)



import json
from shexer.model.statement import Statement
from shexer.model.shape import Shape
from shexer.utils.shapes import build_shapes_name_for_class_uri
from shexer.utils.target_elements import determine_original_target_nodes_if_needed


class ClassShexer(object):

    def __init__(self, class_counts_dict, class_profile_dict=None, class_profile_json_file=None,
                 remove_empty_shapes=True, original_target_classes=None, original_shape_map=None):
        self._class_counts_dict = class_counts_dict
        self._class_profile_dict = class_profile_dict if class_profile_dict is not None else self._load_class_profile_dict_from_file(
            class_profile_json_file)
        self._shapes_list = []
        self._remove_empty_shapes = remove_empty_shapes
        self._original_target_nodes = determine_original_target_nodes_if_needed(remove_empty_shapes=remove_empty_shapes,
                                                                                original_target_classes=original_target_classes,
                                                                                original_shape_map=original_shape_map)

    def shex_classes(self, acceptance_threshold=0):
        self._build_shapes(acceptance_threshold)
        self._clean_shapes()
        self._sort_shapes()

        return self._shapes_list

    def _build_shapes(self, acceptance_threshold):
        for a_class_key in self._class_profile_dict:
            name = build_shapes_name_for_class_uri(a_class_key)
            number_of_instances = float(self._class_counts_dict[a_class_key])
            statements = []
            for a_prop_key in self._class_profile_dict[a_class_key]:
                for a_type_key in self._class_profile_dict[a_class_key][a_prop_key]:
                    for a_cardinality in self._class_profile_dict[a_class_key][a_prop_key][a_type_key]:
                        frequency = self._compute_frequency(number_of_instances,
                                                            self._class_profile_dict
                                                            [a_class_key]
                                                            [a_prop_key]
                                                            [a_type_key]
                                                            [a_cardinality])
                        if frequency >= acceptance_threshold:
                            statements.append(Statement(st_property=a_prop_key,
                                                        st_type=a_type_key,
                                                        cardinality=a_cardinality,
                                                        probability=frequency))

            a_shape = Shape(name=name,
                            class_uri=a_class_key,
                            statements=statements)
            self._shapes_list.append(a_shape)

    def _sort_shapes(self):
        for a_shape in self._shapes_list:
            a_shape.sort_statements(reverse=True,
                                    callback=self._value_to_compare_statements)

    def _clean_shapes(self):

        if not self._remove_empty_shapes:
            return
        shapes_to_remove = self._detect_shapes_to_remove()

        while (len(shapes_to_remove) != 0):
            self._iteration_remove_empty_shapes(shapes_to_remove)
            shapes_to_remove = self._detect_shapes_to_remove()

    def _detect_shapes_to_remove(self):
        result = set()
        for a_shape in self._shapes_list:
            if a_shape.n_statements == 0:
                result.add(a_shape.class_uri)
        return result

    def _iteration_remove_empty_shapes(self, shape_names_to_remove):
        self._remove_shapes_without_statements(shape_names_to_remove)
        self._remove_statements_to_gone_shapes(shape_names_to_remove)

    def _remove_statements_to_gone_shapes(self, shape_names_to_remove):
        for a_shape in self._shapes_list:
            new_statements = []
            for a_statement in a_shape.statements:
                if not a_statement.st_type in shape_names_to_remove:
                    new_statements.append(a_statement)
            a_shape.statements = new_statements

    def _remove_shapes_without_statements(self, shape_names_to_remove):
        new_shape_list = []
        for a_shape in self._shapes_list:
            if not a_shape.name in shape_names_to_remove:
                new_shape_list.append(a_shape)
        self._shapes_list = new_shape_list

    def _value_to_compare_statements(self, a_statement):
        return a_statement.probability

    def _compute_frequency(self, number_of_instances, n_ocurrences_statement):
        return float(n_ocurrences_statement) / number_of_instances

    @staticmethod
    def _load_class_profile_dict_from_file(source_file):
        with open(source_file, "r") as in_stream:
            return json.load(in_stream)



from shexer.consts import SHEX
from shexer.io.shex.formater.shex_serializer import ShexSerializer


def get_shape_serializer(output_format, shapes_list, target_file=None, string_return=False, namespaces_dict=None,
                         tolerance_to_keep_similar_rules=0.15, keep_less_specific=True,
                         instantiation_property=None, discard_useless_constraints_with_positive_closure=True,
                         all_compliant_mode=True, disable_comments=False):
    if output_format == SHEX:
        return ShexSerializer(target_file=target_file,
                              shapes_list=shapes_list,
                              namespaces_dict=namespaces_dict,
                              tolerance_to_keep_similar_rules=tolerance_to_keep_similar_rules,
                              keep_less_specific=keep_less_specific,
                              string_return=string_return,
                              instantiation_property_str=instantiation_property,
                              discard_useless_positive_closures=discard_useless_constraints_with_positive_closure,
                              all_compliant_mode=all_compliant_mode,
                              disable_comments=disable_comments)
    else:
        raise ValueError("Currently unsupported format: " + output_format)



from shexer.core.class_profiler import RDF_TYPE_STR
from shexer.model.IRI import IRI_ELEM_TYPE
from shexer.model.property import Property
from shexer.utils.uri import remove_corners
from shexer.model.fixed_prop_choice_statement import FixedPropChoiceStatement
from shexer.io.shex.formater.consts import SPACES_LEVEL_INDENTATION, POSITIVE_CLOSURE, KLEENE_CLOSURE
from shexer.io.shex.formater.statement_serializers.base_statement_serializer import BaseStatementSerializer  # TODO: REPFACTOR
from shexer.io.shex.formater.statement_serializers.fixed_prop_choice_statement_serializer import FixedPropChoiceStatementSerializer  # TODO: REPFACTOR


class ShexSerializer(object):

    def __init__(self, target_file, shapes_list, namespaces_dict=None,
                 tolerance_to_keep_similar_rules=0.01, keep_less_specific=True, string_return=False,
                 instantiation_property_str=RDF_TYPE_STR, discard_useless_positive_closures=True,
                 all_compliant_mode=True, disable_comments=False):
        self._target_file = target_file
        self._shapes_list = shapes_list
        # self._aceptance_theshold = aceptance_threshold
        self._lines_buffer = []
        self._tolerance = tolerance_to_keep_similar_rules
        self._namespaces_dict = namespaces_dict if namespaces_dict is not None else {}
        self._keep_less_specific = keep_less_specific
        self._string_return = string_return
        self._string_result = ""
        self._instantiation_property_str = self._decide_instantiation_property(instantiation_property_str)
        self._discard_useless_positive_closures = discard_useless_positive_closures
        self._all_compliant_mode = all_compliant_mode
        self._disable_comments = disable_comments

    def serialize_shex(self):

        self._reset_target_file()
        self._serialize_namespaces()
        for a_shape in self._shapes_list:
            self._serialize_shape(a_shape)
        self._flush()
        if self._string_return:
            return self._string_result

    @staticmethod
    def _decide_instantiation_property(instantiation_property_str):
        if instantiation_property_str == None:
            return RDF_TYPE_STR
        if type(instantiation_property_str) == Property:
            return str(instantiation_property_str)
        if type(instantiation_property_str) == str:
            return remove_corners(a_uri=instantiation_property_str,
                                  raise_error_if_no_corners=False)
        raise ValueError("Unrecognized param type to define instantiation property")

    def _serialize_namespaces(self):
        for a_namespace in self._namespaces_dict:
            self._write_line(self._prefix_line(a_namespace), 0)
        self._serialize_empty_namespace()
        self._write_line("", 0)

    def _prefix_line(self, namespace_key):
        return "PREFIX " +  self._namespaces_dict[namespace_key] + ": <" + namespace_key + ">"

    def _serialize_empty_namespace(self):
        self._write_line("PREFIX : <http://weso.es/shapes/>")

    def _serialize_shape(self, a_shape):
        self._serialize_shape_name(a_shape)
        self._serialize_opening_of_rules()
        self._serialize_shape_rules(a_shape)
        self._serialize_closure_of_rule()
        self._serialize_shape_gap()

    def _flush(self):
        self._write_lines_buffer()

    def _write_line(self, a_line, indent_level=0):
        self._lines_buffer.append(self._indentation_spaces(indent_level) + a_line + "\n")
        if len(self._lines_buffer) >= 5000:
            self._write_lines_buffer()
            self._lines_buffer = []

    def _reset_target_file(self):
        if self._string_return:
            return
        with open(self._target_file, "w") as out_stream:
            out_stream.write("")  # Is this necessary? maybe enough to open it in 'w' mode?

    def _write_lines_buffer(self):
        if self._string_return:
            self._string_result += "".join(self._lines_buffer)
        else:
            with open(self._target_file, "a") as out_stream:
                for a_line in self._lines_buffer:
                    out_stream.write(a_line)

    def _indentation_spaces(self, indent_level):
        result = ""
        for i in range(0, indent_level):
            result += SPACES_LEVEL_INDENTATION
        return result

    def _serialize_shape_rules(self, a_shape):
        valid_statements = self._select_valid_statements_of_shape(a_shape)
        if len(valid_statements) == 0:
            return

        if self._all_compliant_mode:
            self._modify_cardinalities_of_statements_non_compliant_with_all_instances(valid_statements)

        if self._disable_comments:
            self._remove_comments_from_statements(valid_statements)

        self._write_valid_statements_of_a_shape(valid_statements)


    def _remove_comments_from_statements(self, valid_statements):
        for a_statement in valid_statements:
            a_statement.remove_comments()

    def _modify_cardinalities_of_statements_non_compliant_with_all_instances(self, statements):
        for a_statement in statements:
            if a_statement.probability != 1:
                self._change_statement_cardinality_to_kleene_closure(a_statement)


    def _change_statement_cardinality_to_kleene_closure(self, statement):
        comment_for_current_sentence = self._turn_statement_into_comment(statement)
        statement.add_comment(comment=comment_for_current_sentence,
                              insert_first=True)
        statement.cardinality = KLEENE_CLOSURE
        statement.probability = 1



    def _write_valid_statements_of_a_shape(self, statements):
        for i in range(0, len(statements) - 1):
            for line_indent_tuple in statements[i]. \
                    get_tuples_to_serialize_line_indent_level(is_last_statement_of_shape=False,
                                                              namespaces_dict=self._namespaces_dict):
                self._write_line(a_line=line_indent_tuple[0],
                                 indent_level=line_indent_tuple[1])
        for line_indent_tuple in statements[len(statements) - 1]. \
                get_tuples_to_serialize_line_indent_level(is_last_statement_of_shape=True,
                                                          namespaces_dict=self._namespaces_dict):
            self._write_line(a_line=line_indent_tuple[0],
                             indent_level=line_indent_tuple[1])


    def _select_valid_statements_of_shape(self, a_shape):
        original_statements = [a_statement for a_statement in a_shape.yield_statements()]
        if len(original_statements) == 0:
            return []

        for a_statement in original_statements:  # TODO Refactor!!! This is not the place to set the serializer
            self._set_serializer_object_for_statements(a_statement)

        result = []
        for i in range(0, len(original_statements)):
            result.append(original_statements[i])

        result = self._group_constraints_with_same_prop_and_obj(result)
        result = self._group_IRI_constraints(result)
        # result = self._group_constraints_with_same_prop_but_different_obj(result)

        result.sort(reverse=True, key=lambda x: x.probability)  # Restoring order completly
        return result

    def _set_serializer_object_for_statements(self, statement):
        statement.serializer_object = BaseStatementSerializer(
            instantiation_property_str=self._instantiation_property_str,
            disable_comments=self._disable_comments)


    def _group_IRI_constraints(self, candidate_statements):
        result = []
        already_visited = set()
        for i in range(0, len(candidate_statements)):
            a_statement = candidate_statements[i]
            if a_statement.st_property == self._instantiation_property_str:
                result.append(a_statement)
                already_visited.add(a_statement)
            else:  # a_statement.st_property != self._instantiation_property_str:
                if a_statement not in already_visited:
                    already_visited.add(a_statement)
                    group_to_decide = [a_statement]

                    for j in range(i + 1, len(candidate_statements)):
                        if self._statements_have_same_prop(a_statement,
                                                           candidate_statements[j]):
                            group_to_decide.append(candidate_statements[j])
                            already_visited.add(candidate_statements[j])
                    if len(group_to_decide) == 1:
                        result.append(a_statement)
                    elif not self._group_contains_IRI_statements(group_to_decide):
                        for a_statement in group_to_decide:
                            result.append(a_statement)
                    else:
                        # pass
                        for a_new_statement in self._compose_statements_with_IRI_objects(group_to_decide):
                            result.append(a_new_statement)
        return result

    def _group_constraints_with_same_prop_and_obj(self, candidate_statements):
        result = []
        already_visited = set()
        for i in range(0, len(candidate_statements)):
            a_statement = candidate_statements[i]
            if a_statement not in already_visited:
                already_visited.add(a_statement)
                group_to_decide = [a_statement]
                for j in range(i + 1, len(candidate_statements)):
                    if self._statements_have_same_tokens(a_statement,
                                                         candidate_statements[j]):
                        group_to_decide.append(candidate_statements[j])
                        already_visited.add(candidate_statements[j])
                if len(group_to_decide) == 1:
                    result.append(a_statement)
                else:
                    result.append(self._decide_best_statement_with_cardinalities_in_comments(group_to_decide))
        return result

    # def _filter_too_similar_statements(self, candidate_statements):
    #     result = []
    #     already_evaluated = set()
    #     for i in range(0, len(candidate_statements)):
    #         if candidate_statements[i] not in already_evaluated:
    #             swapping_statements = []
    #             for j in range(i+1, len(candidate_statements)):
    #                 if self._statements_have_similar_probability(candidate_statements[i],
    #                                                             candidate_statements[j]):
    #                     if self._statements_have_same_tokens(candidate_statements[i],
    #                                                     candidate_statements[j]):
    #                         swapping_statements.append(candidate_statements[j])
    #                         already_evaluated.add(candidate_statements[i])
    #                         already_evaluated.add(candidate_statements[j])
    #                 else:
    #                     break
    #             if len(swapping_statements) == 0:
    #                 result.append(candidate_statements[i])
    #             else:
    #                 result.append(self._decide_best_statement_with_cardinalities_in_comments(swapping_statements + [candidate_statements[i]]))
    #     return result

    def _group_contains_IRI_statements(self, list_of_candidate_statements):
        for a_statement in list_of_candidate_statements:
            if a_statement.st_type == IRI_ELEM_TYPE:
                return True
        return False

    def _is_an_IRI(self, statement_type):
        return statement_type == IRI_ELEM_TYPE or statement_type.startswith("@")  # TODO careful here. Refactor

    def _compose_statements_with_IRI_objects(self, list_of_candidate_statements):
        result = []
        to_compose = []
        for a_statement in list_of_candidate_statements:
            if self._is_an_IRI(a_statement.st_type):
                to_compose.append(a_statement)
            else:
                result.append(a_statement)
        to_compose.sort(reverse=True, key=lambda x: x.probability)
        target_probability = self._get_probability_or_IRI_statement_in_group(to_compose)
        self._remove_IRI_statements_if_useles(to_compose)
        if len(to_compose) > 1:  # There are som sentences to join in an OR
            composed_statement = FixedPropChoiceStatement(st_property=to_compose[0].st_property,
                                                          st_types=[a_statement.st_type for a_statement in to_compose],
                                                          cardinality=POSITIVE_CLOSURE,
                                                          probability=target_probability,
                                                          serializer_object=FixedPropChoiceStatementSerializer(
                                                              instantiation_property_str=self._instantiation_property_str,
                                                              disable_comments=self._disable_comments)
                                                          )
            for a_statement in to_compose:
                if a_statement.st_type != IRI_ELEM_TYPE:
                    composed_statement.add_comment(self._turn_statement_into_comment(a_statement))
            result.append(composed_statement)
        elif len(to_compose) == 1:  # There is just one sentence in the group to join with OR
            result.append(to_compose[0])
        # else  # No sentences to join

        return result


    def _get_probability_or_IRI_statement_in_group(self, group_of_statements):
        for a_statement in group_of_statements:
            if a_statement.st_type == IRI_ELEM_TYPE:
                return a_statement.probability
        raise ValueError("There is no IRI statement within the received group")

    def _remove_IRI_statements_if_useles(self, group_of_statements):
        # I am assuming a group of statements sorted by probability as param
        if len(group_of_statements) <= 1:
            return
        index_of_IRI_statement = -1
        for i in range(0, len(group_of_statements)):
            if group_of_statements[i].st_type == IRI_ELEM_TYPE:
                index_of_IRI_statement = i
                break
        if index_of_IRI_statement != -1:
            if group_of_statements[1].probability == group_of_statements[index_of_IRI_statement].probability:
                # the previous 'if' works, trust me, im engineer
                del group_of_statements[index_of_IRI_statement]

    # def _compose_statement_with_objects_in_or(self, list_of_candidate_statements):
    #     list_of_candidate_statements.sort(reverse=True, key=lambda x: x.probability)
    #     composed_type = list_of_candidate_statements[0].st_type
    #     composed_probability = list_of_candidate_statements[0].probability
    #     for a_sentence in list_of_candidate_statements[1:]:
    #         composed_type += " OR " + a_sentence.st_type
    #         composed_probability += a_sentence.probability
    #
    #
    #
    #     result = Statement(st_property=list_of_candidate_statements[0].st_property,
    #                        st_type=composed_type,
    #                        probability=composed_probability,
    #                        cardinality=POSITIVE_CLOSURE)
    #
    #     for a_sentence in list_of_candidate_statements:
    #         for a_comment in a_sentence.comments:
    #             result.add_comment(a_comment)
    #             result.add_comment(self._turn_statement_into_comment(a_sentence))
    #
    #     return result

    def _decide_best_statement_with_cardinalities_in_comments(self, list_of_candidate_statements):
        if self._discard_useless_positive_closures:
            if self._is_a_group_of_statements_with_useless_positive_closure(list_of_candidate_statements):
                return self._statement_for_a_group_with_a_useless_positive_closure(list_of_candidate_statements)
        list_of_candidate_statements.sort(reverse=True, key=lambda x: x.probability)
        result = None
        if self._keep_less_specific:
            for a_statement in list_of_candidate_statements:
                if a_statement.cardinality == POSITIVE_CLOSURE:
                    result = a_statement
                    break
            if result is None:
                result = list_of_candidate_statements[0]
        else:
            for a_statement in list_of_candidate_statements:
                if a_statement.cardinality != POSITIVE_CLOSURE:
                    result = a_statement
                    break
            if result is None:
                result = list_of_candidate_statements[0]

        for a_statement in list_of_candidate_statements:
            if a_statement.cardinality != result.cardinality:
                result.add_comment(self._turn_statement_into_comment(a_statement))
        return result

    def _is_a_group_of_statements_with_useless_positive_closure(self, list_of_candidate_sentences):
        if len(list_of_candidate_sentences) != 2:
            return False
        if abs(list_of_candidate_sentences[0].probability - list_of_candidate_sentences[1].probability) > self._tolerance:
            return False
        one_if_there_is_a_single_positive_closure = -1
        for a_statement in list_of_candidate_sentences:
            if POSITIVE_CLOSURE == a_statement.cardinality:
                one_if_there_is_a_single_positive_closure *= -1
        if one_if_there_is_a_single_positive_closure == 1:
            return True
        return False

    def _statement_for_a_group_with_a_useless_positive_closure(self, group_of_candidate_statements):
        for a_statement in group_of_candidate_statements:
            if a_statement.cardinality != POSITIVE_CLOSURE:
                return a_statement
        raise ValueError("The received group does not contain any statement with positive closure")



    def _turn_statement_into_comment(self, a_statement):
        return a_statement.comment_representation(namespaces_dict=self._namespaces_dict)

    def _statements_have_similar_probability(self, more_probable_st, less_probable_st):
        if 1.0 - (less_probable_st.probability / more_probable_st.probability) <= self._tolerance:
            return True
        return False

    def _statements_have_same_prop(self, st1, st2):
        if st1.st_property == st2.st_property:
            return True
        return False

    def _statements_have_same_tokens(self, st1, st2):
        if st1.st_property == st2.st_property and st1.st_type == st2.st_type:
            return True
        return False

    def _serialize_shape_name(self, a_shape):
        name = a_shape.name.replace("@", ":")
        if ":" not in name:
            name = "<" + name + ">"
        self._write_line(name)

    def _serialize_opening_of_rules(self):
        self._write_line("{")

    def _serialize_closure_of_rule(self):
        self._write_line("}")

    def _serialize_shape_gap(self):
        self._write_line("")
        self._write_line("")



from shexer.shaper import Shaper
from shexer.consts import TURTLE

a_graph = """
@base <http://library.edu/> . 
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> . 
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> . 
@prefix bf: <http://id.loc.gov/ontologies/bibframe.rdf/> . 

<Instance_1> a bf:Instance ; 
    bf:seriesStatement "Department of State publication" ; 
    bf:seriesEnumeration "8961" ;
    bf:subseriesStatement "General foreign policy series,  9876-5432" ;
    bf:subseriesEnumeration   "volume 310" .
    
bf:Instance_2 a bf:Instance ; 
    bf:seriesStatement "Department of another stuff" ; 
    bf:seriesEnumeration "8961" ;
    rdfs:label "Cosa wena" ;
    bf:subseriesStatement "General foreign policy series,  1232-5674" ;
    bf:subseriesEnumeration   "volume 311" .

bf:Instance rdfs:subClassOf bf:Concept .
bf:Concept rdfs:subClassOf bf:Class .
"""


namespaces_dict={
"http://id.loc.gov/ontologies/bibframe.rdf/" : "bf",
"http://weso.es/" : "",
"http://www.w3.org/2000/01/rdf-schema#": "rdfs"
}

### raw + all_classes

shaper = Shaper(raw_graph=a_graph,
                all_classes_mode=True,
                input_format=TURTLE,
                namespaces_dict=namespaces_dict)

result = shaper.shex_graph(acceptance_threshold=0.5,
                           string_output=True)

print(result)


### raw + target_classes

print("---------")

target_classes = ["http://id.loc.gov/ontologies/bibframe.rdf/Instance"]

shaper = Shaper(raw_graph=a_graph,
                all_classes_mode=False,
                target_classes=target_classes,
                input_format=TURTLE,
                namespaces_dict=namespaces_dict)

result = shaper.shex_graph(acceptance_threshold=0.5,
                           string_output=True)

print(result)

### raw + selectores
from shexer.consts import FIXED_SHAPE_MAP

raw_shape_map = """

{FOCUS rdfs:subClassOf _}@:a_child
bf:Instance_2@:certain_instance
"""
print("---------")
shaper = Shaper(raw_graph=a_graph,
                all_classes_mode=False,
                target_classes=None,
                input_format=TURTLE,
                shape_map_raw=raw_shape_map,
                shape_map_format=FIXED_SHAPE_MAP,
                namespaces_dict=namespaces_dict)

result = shaper.shex_graph(acceptance_threshold=0.5,
                           string_output=True)

print(result)


### raw + selectors + all_classes

print("---------")
shaper = Shaper(raw_graph=a_graph,
                all_classes_mode=True,
                target_classes=None,
                input_format=TURTLE,
                shape_map_raw=raw_shape_map,
                shape_map_format=FIXED_SHAPE_MAP,
                namespaces_dict=namespaces_dict)

result = shaper.shex_graph(acceptance_threshold=0.5,
                           string_output=True)

print(result)


###################################################  File input

print("------------------------------------------------------------"
      "")

file_graph_name = "files\\test_shexer_graph.ttl"

### raw + all_classes

shaper = Shaper(graph_file_input=file_graph_name,
                all_classes_mode=True,
                input_format=TURTLE,
                namespaces_dict=namespaces_dict)

result = shaper.shex_graph(acceptance_threshold=0.5,
                           string_output=True)

print(result)


### raw + target_classes

print("---------")

target_classes = ["http://id.loc.gov/ontologies/bibframe.rdf/Instance"]

shaper = Shaper(graph_file_input=file_graph_name,
                all_classes_mode=False,
                target_classes=target_classes,
                input_format=TURTLE,
                namespaces_dict=namespaces_dict)

result = shaper.shex_graph(acceptance_threshold=0.5,
                           string_output=True)

print(result)

### raw + selectores
from shexer.consts import FIXED_SHAPE_MAP

raw_shape_map = """

{FOCUS rdfs:subClassOf _}@:a_child
bf:Instance_2@:certain_instance
"""
print("---------")
shaper = Shaper(graph_file_input=file_graph_name,
                all_classes_mode=False,
                target_classes=None,
                input_format=TURTLE,
                shape_map_raw=raw_shape_map,
                shape_map_format=FIXED_SHAPE_MAP,
                namespaces_dict=namespaces_dict)

result = shaper.shex_graph(acceptance_threshold=0.5,
                           string_output=True)

print(result)


### raw + selectors + all_classes

print("---------")
shaper = Shaper(graph_file_input=file_graph_name,
                all_classes_mode=True,
                target_classes=None,
                input_format=TURTLE,
                shape_map_raw=raw_shape_map,
                shape_map_format=FIXED_SHAPE_MAP,
                namespaces_dict=namespaces_dict)

result = shaper.shex_graph(acceptance_threshold=0.5,
                           string_output=True)

print(result)


######################################################### endpoint

### endpoint + target
print("---------")

target_classes = ["http://www.wikidata.org/entity/Q44062313", "http://www.wikidata.org/entity/Q54856362"]
endpoint_url = "https://query.wikidata.org/bigdata/namespace/wdq/sparql"
instantiation_property = "http://www.wikidata.org/prop/direct/P1344"

shaper = Shaper(all_classes_mode=False,
                target_classes=target_classes,
                url_endpoint=endpoint_url,
                namespaces_dict=namespaces_dict,
                instantiation_property=instantiation_property,
                track_classes_for_entities_at_last_depth_level=False)

result = shaper.shex_graph(acceptance_threshold=0.5,
                           string_output=True)

print(result)


### endpoint + selectors
endpoint_url = "https://query.wikidata.org/bigdata/namespace/wdq/sparql"
instantiation_property = "http://www.wikidata.org/prop/direct/P31"

raw_shape_map = """

 SPARQL 'SELECT ?s WHERE { ?s <http://www.wikidata.org/prop/direct/P1344> <http://www.wikidata.org/entity/Q44062313> ; <http://www.wikidata.org/prop/direct/P19> <http://www.wikidata.org/entity/Q14317> . }'@:wikifreakoviedo

"""

shaper = Shaper(all_classes_mode=False,
                shape_map_raw=raw_shape_map,
                url_endpoint=endpoint_url,
                namespaces_dict=namespaces_dict,
                instantiation_property=instantiation_property,
                track_classes_for_entities_at_last_depth_level=False)

result = shaper.shex_graph(acceptance_threshold=0.5,
                           string_output=True)
print(result)

## endpoint + selectors + all_clases_mode

endpoint_url = "https://query.wikidata.org/bigdata/namespace/wdq/sparql"
instantiation_property = "http://www.wikidata.org/prop/direct/P31"

raw_shape_map = """

 SPARQL 'SELECT ?s WHERE { ?s <http://www.wikidata.org/prop/direct/P1344> <http://www.wikidata.org/entity/Q44062313> ; <http://www.wikidata.org/prop/direct/P19> <http://www.wikidata.org/entity/Q14317> . }'@:wikifreakoviedo

"""

shaper = Shaper(all_classes_mode=True,
                shape_map_raw=raw_shape_map,
                url_endpoint=endpoint_url,
                namespaces_dict=namespaces_dict,
                instantiation_property=instantiation_property,
                track_classes_for_entities_at_last_depth_level=False)

result = shaper.shex_graph(acceptance_threshold=0.5,
                           string_output=True)
print(result)


######################################################  input_url

### remote url + all_clases

from shexer.consts import RDF_TYPE, RDF_XML

print("---------")
remote_graph_url = "http://xmlns.com/foaf/spec/index.rdf"
instantiation_property = RDF_TYPE

shaper = Shaper(all_classes_mode=True,
                input_format=RDF_XML,
                url_graph_input=remote_graph_url,
                namespaces_dict=namespaces_dict,
                instantiation_property=instantiation_property)

result = shaper.shex_graph(acceptance_threshold=0.5,
                           string_output=True)
print(result)

### remote_url + target_classes

print("---------")
remote_graph_url = "http://xmlns.com/foaf/spec/index.rdf"
instantiation_property = RDF_TYPE

shaper = Shaper(all_classes_mode=False,
                target_classes=["http://www.w3.org/2002/07/owl#AnnotationProperty" ],
                input_format=RDF_XML,
                url_graph_input=remote_graph_url,
                namespaces_dict=namespaces_dict,
                instantiation_property=instantiation_property)

result = shaper.shex_graph(acceptance_threshold=0.5,
                           string_output=True)
print(result)


### remote_url + selectors

print("---------")
remote_graph_url = "http://xmlns.com/foaf/spec/index.rdf"
instantiation_property = RDF_TYPE

raw_selector = """
{FOCUS  rdfs:isDefinedBy _}@<:soyDenifidoPor>
<http://xmlns.com/foaf/0.1/Project>@<:Proyectico>
"""

shaper = Shaper(all_classes_mode=False,
                shape_map_raw=raw_selector,
                input_format=RDF_XML,
                url_graph_input=remote_graph_url,
                namespaces_dict=namespaces_dict,
                instantiation_property=instantiation_property)

result = shaper.shex_graph(acceptance_threshold=0.5,
                           string_output=True)
print(result)


### remote_url + selectors + all_classes

print("---------")
remote_graph_url = "http://xmlns.com/foaf/spec/index.rdf"
instantiation_property = RDF_TYPE

raw_selector = """
{FOCUS  rdfs:isDefinedBy _}@<:soyDenifidoPor>
<http://xmlns.com/foaf/0.1/Project>@<:Proyectico>
"""

shaper = Shaper(all_classes_mode=True,
                shape_map_raw=raw_selector,
                input_format=RDF_XML,
                url_graph_input=remote_graph_url,
                namespaces_dict=namespaces_dict,
                instantiation_property=instantiation_property)

result = shaper.shex_graph(acceptance_threshold=0.5,
                           string_output=True)
print(result)






raw_content = """
{
    "@id": "http://store.example.com/",
    "@type": "Store",
    "name": "Links Bike Shop",
    "description": "The most linked bike store on earth!",
    "product": [
        {
            "@id": "p:links-swift-chain",
            "@type": "Product",
            "name": "Links Swift Chain",
            "description": "A fine chain with many links.",
            "category": ["cat:parts", "cat:chains"],
            "price": "10.00",
            "stock": 10
        },
        {
            "@id": "p:links-speedy-lube",
            "@type": "Product",
            "name": "Links Speedy Lube",
            "description": "Lubricant for your chain links.",
            "category": ["cat:lube", "cat:chains"],
            "price": "5.00",
            "stock": 20
        }
    ],
    "@context": {
        "Store": "http://ns.example.com/store#Store",
        "Product": "http://ns.example.com/store#Product",
        "product": "http://ns.example.com/store#product",
        "category":
        {
          "@id": "http://ns.example.com/store#category",
          "@type": "@id"
        },
        "price": "http://ns.example.com/store#price",
        "stock": "http://ns.example.com/store#stock",
        "name": "http://purl.org/dc/terms/title",
        "description": "http://purl.org/dc/terms/description",
        "p": "http://store.example.com/products/",
        "cat": "http://store.example.com/category/"
    }
}
"""

from shexer.shaper import Shaper
from shexer.consts import JSON_LD

shaper = Shaper(all_classes_mode=True,
                raw_graph=raw_content,
                input_format=JSON_LD)

result = shaper.shex_graph(string_output=True, acceptance_threshold=0)
print(result)



from shexer.utils.factories.triple_yielders_factory import produce_shape_map_according_to_input


def get_shape_map_if_needed(sm_format, sgraph, namespaces_prefix_dict, target_classes,
                            file_target_classes, shape_map_file, shape_map_raw,
                            instantiation_property, shape_map_already_built=None):
    if shape_map_file is None and shape_map_raw is None:
        return None
    return produce_shape_map_according_to_input(sm_format=sm_format,
                                                sgraph=sgraph,
                                                namespaces_prefix_dict=namespaces_prefix_dict,
                                                target_classes=target_classes,
                                                file_target_classes=file_target_classes,
                                                shape_map_file=shape_map_file,
                                                shape_map_raw=shape_map_raw,
                                                instantiation_property=instantiation_property,
                                                shape_map_already_built=shape_map_already_built)


from shexer.model.property import Property
from shexer.utils.uri import remove_corners
from shexer.utils.factories.h_tree import get_basic_h_tree
from shexer.core.instances.annotators.annotator_func import get_proper_annotator
from shexer.core.instances.abstract_instance_tracker import AbstractInstanceTracker



_RDF_TYPE = Property(content="http://www.w3.org/1999/02/22-rdf-syntax-ns#type")
_RDFS_SUBCLASS_OF = Property(content="http://www.w3.org/2000/01/rdf-schema#subClassOf")


class InstanceTracker(AbstractInstanceTracker):

    def __init__(self, target_classes, triples_yielder, instantiation_property=_RDF_TYPE,
                 all_classes_mode=False, subclass_property=_RDFS_SUBCLASS_OF, track_hierarchies=True,
                 shape_qualifiers_mode=False, namespaces_for_qualifier_props=None):
        self._target_classes = target_classes
        self._all_classes_mode = all_classes_mode
        self._instances_dict = self._build_instances_dict(self._target_classes)
        self._triples_yielder = triples_yielder
        self._instantiation_property = self._decide_instantiation_property(instantiation_property)
        self._relevant_triples = 0
        self._not_relevant_triples = 0
        self._subclass_property = subclass_property
        self._track_hierarchies = track_hierarchies
        self._shape_qualifiers_mode = shape_qualifiers_mode
        self._namespaces_for_qualifiers_props = [] if namespaces_for_qualifier_props is None else namespaces_for_qualifier_props

        self._htree = get_basic_h_tree() if track_hierarchies else None
        self._classes_considered_in_htree = set() if track_hierarchies else None

        self._annotator = get_proper_annotator(track_hierarchies=track_hierarchies,
                                               instance_tracker_ref=self)

    @property
    def disambiguator_prefix(self):
        return "class_"

    @property
    def relevant_triples(self):
        return self._relevant_triples

    @property
    def not_relevant_triples(self):
        return self._not_relevant_triples

    @property
    def htree(self):
        return self._htree

    def track_instances(self):
        self._reset_count()
        for a_revelant_triple in self._yield_relevant_triples():
            self._annotator.annotate_triple(a_revelant_triple)
        self._annotator.annotation_post_parsing()

        return self._instances_dict

    def _yield_relevant_triples(self):
        for a_triple in self._triples_yielder.yield_triples():
            if self._annotator.is_relevant_triple(a_triple):
                self._relevant_triples += 1
                yield a_triple
            else:
                self._not_relevant_triples += 1

    def _reset_count(self):
        self._relevant_triples = 0
        self._not_relevant_triples = 0
        self._htree = get_basic_h_tree()

    def is_an_instantiation_prop(self, a_property):
        return a_property == self._instantiation_property

    def is_a_subclass_property(self, a_property):
        return a_property == self._subclass_property

    @staticmethod
    def _build_instances_dict(target_classes):
        """
        If there are target classes, we can build their dictionary now. Otherwise, all_classes_mode or
        shape_qualifiers_mode will fill the instances dict in future methods

        :param target_classes:
        :return:
        """
        result = {}
        if target_classes is not None:
            for a_class in target_classes:
                result[a_class.iri] = set()
        # return result  # In this case, we will add keys on the fly, while parsing the input graph.

        return result

    @staticmethod
    def _decide_instantiation_property(instantiation_property):
        if instantiation_property == None:
            return _RDF_TYPE
        if type(instantiation_property) == type(_RDF_TYPE):
            return instantiation_property
        if type(instantiation_property) == str:
            return Property(remove_corners(a_uri=instantiation_property,
                                           raise_error_if_no_corners=False))
        raise ValueError("Unrecognized param type to define instantiation property")

    # def _anotate_direct_children_of_IRI(self):
    #     for a_key_class in self._instances_dict:
    #         self._classes_considered_in_htree.add(a_key_class)
    #     iri_node = self._htree.iri_node
    #     for a_key_class in self._classes_considered_in_htree:
    #         a_class_node = self._get_appropiate_iri_node_and_add_to_htree_if_needed(a_key_class)
    #         if not a_class_node.has_parents():
    #             a_class_node.add_parent(iri_node)

    #
    # def _anotate_instance(self, a_triple):
    #     self._instances_dict[a_triple[_O].iri].add(a_triple[_S].iri)

    # def _is_a_relevant_triple(self, a_triple):
    #     """
    #     It returns True if the triple has rdf:type as predicate and one of the target classes as object
    #
    #     :return: bool
    #     """
    #     if self._track_hierarchies and a_triple[_P] == self._subclass_property:
    #         return True
    #     elif a_triple[_P] != self._instantiation_property:
    #         return False
    #     elif self._all_classes_mode:
    #         if a_triple[_O].iri not in self._instances_dict:
    #             self.add_new_class_to_instances_dict(a_triple[_O].iri)
    #             return True
    #     elif a_triple[_O].iri not in self._instances_dict:
    #         return False
    #     return True








# import os
# import io
# import zipfile
#
# base_source_dir = "C:\\Users\\Dani\\Documents\\EII\\coronavirus\\shex\\misterios_andra\\"
# base_dest_dir = "C:\\Users\\Dani\\Documents\\EII\\coronavirus\\shex\\target\\"
#
# import gzip
# f = gzip.open(base_source_dir + 'CC2rdf.xsl.gz', 'rt')
# file_content = f.read()
# with open(base_dest_dir + 'ahiesta.xsl', "w") as out_str:
#     out_str.write(file_content)
# f.close()
#
#
# print("Done!")
#
#
# def extract_folder_files(gz_source_folder, dest_folder):
#     folder = gzip.open(gz_source_folder, 'rt')
#     for files in folder.na
#
# # def extract(filename):
# #     z = zipfile.ZipFile(filename)
# #     for f in z.namelist():
# #         # get directory name from file
# #         dirname = os.path.splitext(f)[0]
# #         # create new directory
# #         os.mkdir(dirname)
# #         # read inner zip file into bytes buffer
# #         content = io.BytesIO(z.read(f))
# #         zip_file = zipfile.ZipFile(content)
# #         for i in zip_file.namelist():
# #             zip_file.extract(i, dirname)

# ! /usr/bin/env python
# -*- coding: UTF-8 -*-

"""A command line utility for recusively extracting nested tar archives."""

__author__ = "Pushpak Dagade"
__date__ = "$4 July, 2011 3:00:00 PM$"

import os
import sys
import re
import tarfile
import gzip
from argparse import ArgumentParser

major_version = 1
minor_version = 1
error_count = 0

file_extensions = ('tar', 'tgz', 'tbz', 'tb2', 'tar.gz', 'tar.bz2')
# Edit this according to the archive types you want to extract. Keep in
# mind that these should be extractable by the tarfile module.

__all__ = ['ExtractNested', 'WalkTreeAndExtract']



def FileExtension(file_name):
    """Return the file extension of file
    'file' should be a string. It can be either the full path of
    the file or just its name (or any string as long it contains
    the file extension.)
    Example #1:
    input (file) -->  'abc.tar.gz'
    return value -->  'tar.gz'

    Example #2:
    input (file) -->  'abc.tar'
    return value -->  'tar'

    """
    match = re.compile(r"^.*?[.](?P<ext>tar[.]gz|tar[.]bz2|\w+)$",
                       re.VERBOSE | re.IGNORECASE).match(file_name)

    if match:  # if match != None:
        ext = match.group('ext')
        return ext
    else:
        return ''  # there is no file extension to file_name


def AppropriateFolderName(folder_fullpath):
    """Return a folder (path) such that it can be safely created in
    without replacing any existing folder in it.
    Check if the folder folder_fullpath exists. If no, return folder_fullpath
    (without changing, because it can be safely created
    without replacing any already existing folder). If yes, append an
    appropriate number to the folder_fullpath such that this new folder_fullpath
    can be safely created.
    Examples:
    folder_name  = '/a/b/untitled folder'
    return value = '/a/b/untitled folder'   (no such folder already exists.)
    folder_name  = '/a/b/untitled folder'
    return value = '/a/b/untitled folder 1' (the folder '/a/b/untitled folder'
                                            already exists but no folder named
                                            '/a/b/untitled folder 1' exists.)
    folder_name  = '/a/b/untitled folder'
    return value = '/a/b/untitled folder 2' (the folders '/a/b/untitled folder'
                                            and '/a/b/untitled folder 1' both
                                            already exist but no folder
                                            '/a/b/untitled folder 2' exists.)

    """
    if os.path.exists(folder_fullpath):
        folder_name = os.path.basename(folder_fullpath)
        parent_fullpath = os.path.dirname(folder_fullpath)
        match = re.compile(r'^(?P<name>.*)[ ](?P<num>\d+)$').match(folder_name)
        if match:  # if match != None:
            name = match.group('name')
            number = match.group('num')
            new_folder_name = '%s %d' % (name, int(number) + 1)
            new_folder_fullpath = os.path.join(parent_fullpath, new_folder_name)
            return AppropriateFolderName(new_folder_fullpath)
            # Recursively call itself so that it can be check whether a
            # folder with path new_folder_fullpath already exists or not.
        else:
            new_folder_name = '%s 1' % folder_name
            new_folder_fullpath = os.path.join(parent_fullpath, new_folder_name)
            return AppropriateFolderName(new_folder_fullpath)
            # Recursively call itself so that it can be check whether a
            # folder with path new_folder_fullpath already exists or not.
    else:
        return folder_fullpath


def Extract(tarfile_fullpath, delete_tar_file=True):
    """Extract the tarfile_fullpath to an appropriate* folder of the same
    name as the tar file (without an extension) and return the path
    of this folder.
    If delete_tar_file is True, it will delete the tar file after
    its extraction; if False, it won`t. Default value is True as you
    would normally want to delete the (nested) tar files after
    extraction. Pass a False, if you don`t want to delete the
    tar file (after its extraction) you are passing.
    """
    try:
        # print
        # "Extracting '%s'" % tarfile_fullpath,
        tar = tarfile.open(tarfile_fullpath)
        extract_folder_fullpath = AppropriateFolderName(tarfile_fullpath[: \
                                                                         -1 * len(FileExtension(tarfile_fullpath)) - 1])
        extract_folder_name = os.path.basename(extract_folder_fullpath)
        # print
        # "to '%s'..." % extract_folder_name,
        tar.extractall(extract_folder_fullpath)
        print("Done!")
        tar.close()
        if delete_tar_file: os.remove(tarfile_fullpath)
        return extract_folder_name

    except Exception:
        # Exceptions can occur while opening a damaged tar file.
        print('(Error)\n(%s)' % str(sys.exc_info()[1]).capitalize())
        global error_count
        error_count += 1


def ExtractGz(tarfile_fullpath, delete_tar_file=True):
    try:
        print("Extracting '%s'" % tarfile_fullpath,)
        gz = gzip.open(tarfile_fullpath)
        extract_folder_fullpath = AppropriateFolderName(tarfile_fullpath[: \
                                                                         -1 * len(FileExtension(tarfile_fullpath)) - 1])
        extract_folder_name = os.path.basename(extract_folder_fullpath)
        print("to '%s'..." % extract_folder_name,)
        output = open( extract_folder_fullpath, "wb")
        output.write(gz.read())
        print("Done!")
        gz.close()
        output.close()
        if delete_tar_file: os.remove(tarfile_fullpath)
        return extract_folder_name

    except Exception:
        # Exceptions can occur while opening a damaged tar file.
        print('(Error)\n(%s)' % str(sys.exc_info()[1]).capitalize())
        global error_count
        error_count += 1


def WalkTreeAndExtract(parent_dir):
    """Recursively descend the directory tree rooted at parent_dir
    and extract each tar file on the way down (recursively)."""
    try:
        dir_contents = os.listdir(parent_dir)
    except OSError:
        # Exception can occur if trying to open some folder whose
        # permissions this program does not have.
        print('Error occured. Could not open folder %s\n%s' \
            % (parent_dir, str(sys.exc_info()[1]).capitalize()) )
        global error_count
        error_count += 1
        return

    for content in dir_contents:
        content_fullpath = os.path.join(parent_dir, content)
        if os.path.isdir(content_fullpath):
            # If content is a folder, walk down it completely.
            WalkTreeAndExtract(content_fullpath)
        elif os.path.isfile(content_fullpath):
            # If content is a file, check if it is a tar file.
            if FileExtension(content_fullpath) in file_extensions:
                # If yes, extract its contents to a new folder.
                extract_folder_name = Extract(content_fullpath)
                if extract_folder_name:  # if extract_folder_name != None:
                    dir_contents.append(extract_folder_name)
                    # Append the newly extracted folder to dir_contents
                    # so that it can be later searched for more tar files
                    # to extract.
            elif FileExtension(content_fullpath) == 'gz':
                # If yes, extract its contents to a new folder.
                extract_folder_name = ExtractGz(content_fullpath)
                if extract_folder_name:  # if extract_folder_name != None:
                    dir_contents.append(extract_folder_name)
                    # Append the newly extracted folder to dir_contents
                    # so that it can be later searched for more tar files
                    # to extract.
        else:
            # Unknown file type.
            print('Skipping %s. <Neither file nor folder>' % content_fullpath)


def ExtractNested(tarfile_fullpath):
    extract_folder_name = Extract(tarfile_fullpath, False)
    if extract_folder_name:  # if extract_folder_name != None
        extract_folder_fullpath = os.path.join(os.path.dirname(
            tarfile_fullpath), extract_folder_name)
        WalkTreeAndExtract(extract_folder_fullpath)
        # Given tar file is extracted to extract_folder_name. Now descend
        # down its directory structure and extract all other tar files
        # (recursively).

def ExtractFolder(folder_name):
    if folder_name:  # if extract_folder_name != None
        # extract_folder_fullpath = os.path.join(os.path.dirname(
        #     folder_name), folder_name)
        WalkTreeAndExtract(folder_name)
        # Given tar file is extracted to extract_folder_name. Now descend
        # down its directory structure and extract all other tar files
        # (recursively).


if __name__ == '__main__':
    # Use a parser for parsing command line arguments
    parser = ArgumentParser(description='Nested tar archive extractor %d.%d' \
                                        % (major_version, minor_version))
    parser.add_argument('tar_paths', metavar='path', type=str, nargs='+',
                        help='Path of the tar file to be extracted.')
    extraction_paths = parser.parse_args().tar_paths

    # Consider each argument passed as a file path and extract it.
    for argument in extraction_paths:
        if os.path.exists(argument):
            # print       # a blank line
            ExtractFolder(argument)
        else:
            print('Not a valid path: %s' % argument)
            error_count += 1
    if error_count != 0: print('%d error(s) occured.' % error_count)



from shexer.model.shape import STARTING_CHAR_FOR_SHAPE_NAME

def build_shapes_name_for_class_uri(class_uri):
    if "#" in class_uri and class_uri[-1] != "#":
        return STARTING_CHAR_FOR_SHAPE_NAME + class_uri[class_uri.rfind("#") + 1:]
    if "/" in class_uri:
        if class_uri[-1] != "/":
            return STARTING_CHAR_FOR_SHAPE_NAME + class_uri[class_uri.rfind("/") + 1:]
        else:
            return STARTING_CHAR_FOR_SHAPE_NAME + class_uri[class_uri[:-1].rfind("/") + 1:]
    else:
        return class_uri


def build_shape_name_for_qualifier_prop_uri(prop_uri):
    result = prop_uri
    if "#" in prop_uri and prop_uri[-1] != "#":
        return STARTING_CHAR_FOR_SHAPE_NAME + prop_uri[prop_uri.rfind("#") + 1:]
    if "/" in prop_uri:
        if prop_uri[-1] != "/":
            return STARTING_CHAR_FOR_SHAPE_NAME + prop_uri[prop_uri.rfind("/") + 1:]
        else:
            return STARTING_CHAR_FOR_SHAPE_NAME + prop_uri[prop_uri[:-1].rfind("/") + 1:]
    else:
        return result.upper()


from shexer.core.class_shexer import ClassShexer


def get_class_shexer(class_instances_target_dict,
                     class_profile_dict,
                     remove_empty_shapes,
                     original_target_classes,
                     original_shape_map):
    class_count_dicts = {}

    for a_class_key in class_instances_target_dict:
        class_count_dicts[a_class_key] = len(class_instances_target_dict[a_class_key])

    return ClassShexer(
        class_counts_dict=class_count_dicts,
        class_profile_dict=class_profile_dict,
        class_profile_json_file=None,
        remove_empty_shapes=remove_empty_shapes,
        original_target_classes=original_target_classes,
        original_shape_map=original_shape_map
    )


list_of_names = [
    # "infobox_properties_mapped_en",
                 "infobox_properties_en",
                 "instance_types_en",
                 "mappingbased_objects_en",
                 "persondata_en",
                 "specific_mappingbased_properties_en",
                 # "topical_concepts_en"
]

template_file = "C:\\Users\\Dani\\Documents\\EII\\doctorado\\datasets\\dbpedia_endpoint\\{}.ttl"

namespaces_dict = {

"http://id.loc.gov/ontologies/bibframe.rdf/" : "bf",
"http://weso.es/" : "",
"http://www.w3.org/2000/01/rdf-schema#": "rdfs",
"http://www.wikidata.org/prop/direct/": "wdt",
"http://www.wikidata.org/prop/": "wd",
    "http://dbpedia.org/ontology/": "dbo",
    "http://dbpedia.org/resource/": "dbr",
    "http://www.w3.org/2001/XMLSchema#" : "xsd",
    "http://dbpedia.org/class/yago/" : "yago",
    "http://www.w3.org/1999/02/22-rdf-syntax-ns#" : "rdf",
    "http://www.w3.org/2003/01/geo/wgs84_pos#" : "geo",
    "http://xmlns.com/foaf/0.1/" : "foaf",
    "http://purl.org/dc/elements/1.1/" : "dc",
    "http://dbpedia.org/resource/Category:" : "dbc",
    "http://dbpedia.org/property/" : "dbp",
    "http://www.w3.org/XML/1998/namespace" : "xml"
}


from shexer.shaper import Shaper
from shexer.consts import NT
import time

print("Starting...")
stime = time.time()
print(stime)

shaper = Shaper(target_classes=["http://dbpedia.org/ontology/Award"],
                graph_list_of_files_input=[template_file.format(a_file) for a_file in list_of_names],
                input_format=NT,
                namespaces_to_ignore=namespaces_dict,
                instances_file_input=template_file.format("instance_types_en"),
                all_instances_are_compliant_mode=True)

print("Graph built...")

result = shaper.shex_graph(string_output=True, acceptance_threshold=0.4)


etime = time.time()
print("Time ", etime - stime)
print(result)




from shexer.model.IRI import IRI_ELEM_TYPE
from shexer.utils.shapes import build_shapes_name_for_class_uri
from shexer.utils.target_elements import determine_original_target_nodes_if_needed
from shexer.model.property import Property
from shexer.model.bnode import BNode
from shexer.utils.uri import remove_corners

RDF_TYPE_STR = "http://www.w3.org/1999/02/22-rdf-syntax-ns#type"


_ONE_TO_MANY = "+"

_S = 0
_P = 1
_O = 2


class ClassProfiler(object):

    def __init__(self, triples_yielder, target_classes_dict, instantiation_property_str=RDF_TYPE_STR,
                 remove_empty_shapes=True, original_target_classes=None, original_shape_map=None ):
        self._triples_yielder = triples_yielder
        self._target_classes_dict = target_classes_dict
        self._instances_shape_dict = {}
        self._classes_shape_dict = self._build_classes_shape_dict_with_just_classes()
        self._shape_names_dict = self._build_shape_names_dict()
        self._relevant_triples = 0
        self._instantiation_property_str = self._decide_instantiation_property(instantiation_property_str)
        self._remove_empty_shapes=remove_empty_shapes
        self._original_target_nodes = determine_original_target_nodes_if_needed(remove_empty_shapes=remove_empty_shapes,
                                                                                original_target_classes=original_target_classes,
                                                                                original_shape_map=original_shape_map)



    def profile_classes(self):
        self._build_shape_of_instances()
        # print("Number of relevant triples", self._relevant_triples)
        # print("Profiler... shape of instances built!")
        self._build_class_profile()
        # print("Profiler... class profile built!")
        self._clean_class_profile()
        return self._classes_shape_dict

    def get_target_classes_dict(self):
        return self._target_classes_dict

    @staticmethod
    def _decide_instantiation_property(instantiation_property_str):
        if instantiation_property_str == None:
            return RDF_TYPE_STR
        if type(instantiation_property_str) == Property:
            return str(instantiation_property_str)
        if type(instantiation_property_str) == str:
            return remove_corners(a_uri=instantiation_property_str,
                                  raise_error_if_no_corners=False)
        raise ValueError("Unrecognized param type to define instantiation property")


    def _build_shape_names_dict(self):
        result = {}
        for a_class in self._target_classes_dict:
            name = build_shapes_name_for_class_uri(a_class)
            result[a_class] = name
        return result


    def _build_classes_shape_dict_with_just_classes(self):
        result = {}
        for a_class_key in self._target_classes_dict:
            result[a_class_key] = {}
        return result


    def _infer_3tuple_features(self, an_instance):
        result = []
        for a_prop in self._instances_shape_dict[an_instance]:
            for a_type in self._instances_shape_dict[an_instance][a_prop]:
                for a_valid_cardinality in self._infer_valid_cardinalities(a_prop,
                                                                           self._instances_shape_dict[an_instance][a_prop][a_type]):
                    result.append( (a_prop, a_type, a_valid_cardinality) )
        return result


    def _infer_valid_cardinalities(self, a_property, a_cardinality):
        """
        Special teratment for self._instantiation_property_str. If thats the property, we are targetting specific URIs
        instead of the type IRI.
        Cardinality will be always "1"
        :param a_property:
        :param a_cardinality:
        :return:
        """
        if a_property == self._instantiation_property_str:
            yield 1
        else:
            yield a_cardinality
            yield _ONE_TO_MANY


    def _build_class_profile(self):
        for an_instance in self._instances_shape_dict:
            feautres_3tuple = self._infer_3tuple_features(an_instance)
            for a_class in self._target_classes_dict:
                if self._is_instance_of_class(an_instance, a_class):
                    self._annotate_instance_features_for_class(a_class, feautres_3tuple)

    def _clean_class_profile(self):
        if not self._remove_empty_shapes:
            return
        shapes_to_remove = self._detect_shapes_to_remove()

        while(len(shapes_to_remove) != 0):
            self._iteration_remove_empty_shapes(shapes_to_remove)
            shapes_to_remove = self._detect_shapes_to_remove()

    def _detect_shapes_to_remove(self):
        shapes_to_remove = set()
        for a_shape_key in self._classes_shape_dict:
            if not self._is_original_target_shape(a_shape_key):
                if not self._has_it_annotated_features(a_shape_key):
                    shapes_to_remove.add(a_shape_key)
        return shapes_to_remove

    def _is_original_target_shape(self, shape_label):
        return shape_label in self._original_target_nodes

    def _has_it_annotated_features(self, shape_label):
        if shape_label not in self._classes_shape_dict:
            return False
        return len(self._classes_shape_dict[shape_label]) > 0

    def _iteration_remove_empty_shapes(self, target_shapes):
        for a_shape_label_key in self._classes_shape_dict:
            for a_prop_key in self._classes_shape_dict[a_shape_label_key]:
                # print(self._classes_shape_dict[a_shape_label_key][a_prop_key])
                for a_shape_to_remove in target_shapes:
                    if a_shape_to_remove in self._classes_shape_dict[a_shape_label_key][a_prop_key]:
                        del self._classes_shape_dict[a_shape_label_key][a_prop_key][a_shape_to_remove]
        for a_shape_to_remove in target_shapes:
            if a_shape_to_remove in self._classes_shape_dict:
                del self._classes_shape_dict[a_shape_to_remove]


    def _annotate_instance_features_for_class(self, a_class, features_3tuple):
        for a_feature_3tuple in features_3tuple:
            self._introduce_needed_elements_in_shape_classes_dict(a_class, a_feature_3tuple)
            # 3tuple: 0->str_prop, 1->str_type, 2->cardinality
            self._classes_shape_dict[a_class][a_feature_3tuple[0]][a_feature_3tuple[1]][a_feature_3tuple[2]] += 1

    def _introduce_needed_elements_in_shape_classes_dict(self, a_class, a_feature_3tuple):
        str_prop = a_feature_3tuple[0]
        str_type = a_feature_3tuple[1]
        cardinality = a_feature_3tuple[2]
        if str_prop not in self._classes_shape_dict[a_class]:
            self._classes_shape_dict[a_class][str_prop] = {}
        if str_type not in self._classes_shape_dict[a_class][str_prop]:
            self._classes_shape_dict[a_class][str_prop][str_type] = {}
        if cardinality not in self._classes_shape_dict[a_class][str_prop][str_type]:
            self._classes_shape_dict[a_class][str_prop][str_type][cardinality] = 0


    def _is_instance_of_class(self, an_instance_str, a_class_str):
        if an_instance_str in self._target_classes_dict[a_class_str]:
            return True
        return False


    def _build_shape_of_instances(self):
        for a_triple in self._yield_relevant_triples():
            self._relevant_triples += 1
            self._annotate_feature_of_target_instance(a_triple)


    def _annotate_feature_of_target_instance(self, a_triple):
        str_subj = a_triple[_S].iri
        str_prop = a_triple[_P].iri
        type_obj = self._decide_type_obj(a_triple[_O], str_prop)

        obj_shapes = [] if type_obj != IRI_ELEM_TYPE else self._decide_shapes_obj(a_triple[_O].iri)

        self._introduce_needed_elements_in_shape_instances_dict(str_subj=str_subj,
                                                                str_prop=str_prop,
                                                                type_obj=type_obj,
                                                                obj_shapes=obj_shapes)

        self._instances_shape_dict[str_subj][str_prop][type_obj] += 1
        for a_shape in obj_shapes:
            self._instances_shape_dict[str_subj][str_prop][a_shape] += 1

    def _decide_type_obj(self, original_object, str_prop):
        """
        Special traetment for self._instantiation_property_str property. We look for ValueSets instead of types when this property appears.

        :param original_object:
        :param str_prop:
        :return:
        """
        if str_prop != self._instantiation_property_str:
            return original_object.elem_type
        return original_object.iri


    def _introduce_needed_elements_in_shape_instances_dict(self, str_subj, str_prop, type_obj, obj_shapes):
        if str_subj not in self._instances_shape_dict:
            self._instances_shape_dict[str_subj] = {}
        if str_prop not in self._instances_shape_dict[str_subj]:
            self._instances_shape_dict[str_subj][str_prop] = {}
        if type_obj not in self._instances_shape_dict[str_subj][str_prop]:
            self._instances_shape_dict[str_subj][str_prop][type_obj] = 0
        for a_shape in obj_shapes:
            if a_shape not in self._instances_shape_dict[str_subj][str_prop]:
                self._instances_shape_dict[str_subj][str_prop][a_shape] = 0

    def _decide_shapes_obj(self, str_obj):
        result = []
        for class_key in self._target_classes_dict:
            if str_obj in self._target_classes_dict[class_key]:
                result.append(self._shape_names_dict[class_key])
        return result


    def _yield_relevant_triples(self):
        for a_triple in self._triples_yielder.yield_triples():
            if self._is_a_relevant_triple(a_triple):
                yield a_triple

    def _is_a_relevant_triple(self, a_triple):
        """
        The subject of the triple must be an instance of at least one of the target classes.
        If it is, it returns True. False in the opposite case

        :param a_triple:
        :return: bool
        """
        return True if self._is_subject_in_target_classes(a_triple) else False

    def _is_subject_in_target_classes(self, a_triple):
        subj = a_triple[_S]
        if isinstance(subj, BNode):
            return False
        str_subj = subj.iri
        for class_key in self._target_classes_dict:
            if str_subj in self._target_classes_dict[class_key]:
                return True
        return False




from shexer.core.instances.abstract_instance_tracker import AbstractInstanceTracker

class MixedInstanceTracker(AbstractInstanceTracker):

    def __init__(self, list_of_instance_trackers):
        self._reference_instance_tracker = list_of_instance_trackers[0]
        self._secondary_instance_trackers = [] if len(list_of_instance_trackers) <= 1 else list_of_instance_trackers[1:]
        self._disambiguator_counter = 0


    def track_instances(self):
        reference_instances_dict = self._reference_instance_tracker.track_instances()
        for a_tracker in self._secondary_instance_trackers:
            self._integrate_dicts(reference_dict=reference_instances_dict,
                                  new_dict=a_tracker.track_instances(),
                                  new_tracker=a_tracker)
        return reference_instances_dict


    def _integrate_dicts(self, reference_dict, new_dict, new_tracker):
        if not self._is_there_key_ambiguity(reference_iterable=reference_dict,
                                            new_iterable=new_dict):
            self._integrate_dicts_with_no_key_ambiguity(reference_dict=reference_dict,
                                                        new_dict=new_dict)
        else:
            self._integrate_dicts_with_key_ambiguity(reference_dict=reference_dict,
                                                     new_dict=new_dict,
                                                     new_tracker=new_tracker)


    def _is_there_key_ambiguity(self, reference_iterable, new_iterable):
        """
        If possible, provide in reference iterable a structure with O(1) for the 'in' operator

        :param reference_iterable:
        :param new_iterable:
        :return:
        """
        for a_new_key in new_iterable:
            if a_new_key in reference_iterable:
                return True
        return False

    def _integrate_dicts_with_no_key_ambiguity(self, reference_dict, new_dict):
        for a_new_key, a_new_value in new_dict.items():
            reference_dict[a_new_key] = a_new_value

    def _integrate_dicts_with_key_ambiguity(self, reference_dict, new_dict, new_tracker):
        prefix = self._get_disambiguator_prefix_label(reference_dict=reference_dict,
                                                      new_dict=new_dict,
                                                      new_tracker=new_tracker)
        for a_new_key, a_new_value in new_dict.items():
            reference_dict[prefix + a_new_key] = a_new_value

    def _get_disambiguator_prefix_label(self, reference_dict, new_dict, new_tracker):
        base_disam = new_tracker.disambiguator_prefix
        current_disam = base_disam
        new_keys = [current_disam + a_new_key for a_new_key in new_dict]
        while self._is_there_key_ambiguity(reference_iterable=reference_dict,
                                           new_iterable=new_keys):
            self._disambiguator_counter += 1
            current_disam = base_disam + str(self._disambiguator_counter)
            new_keys = [current_disam + a_new_key for a_new_key in new_dict]
        return base_disam











from shexer.core.instances.abstract_instance_tracker import AbstractInstanceTracker


class ShapeMapInstanceTracker(AbstractInstanceTracker):

    def __init__(self, shape_map):
        self._shape_map = shape_map
        self._instances_dict = {}

    def track_instances(self):
        for an_item in self._shape_map.yield_items():
            self._solve_targets_of_an_item(an_item)
        return self._instances_dict

    def _solve_targets_of_an_item(self, an_item):
        self._instances_dict[an_item.shape_label] = an_item.node_selector.get_target_nodes()

    @property
    def disambiguator_prefix(self):
        return "custom_"


from shexer.io.shex.formater.statement_serializers.base_statement_serializer import BaseStatementSerializer
from shexer.io.shex.formater.consts import SPACES_GAP_BETWEEN_TOKENS, KLEENE_CLOSURE


class FixedPropChoiceStatementSerializer(BaseStatementSerializer):

    def __init__(self, instantiation_property_str, disable_comments=False):
        super(FixedPropChoiceStatementSerializer, self).__init__(instantiation_property_str=instantiation_property_str,
                                                                 disable_comments=disable_comments)

    def serialize_statement_with_indent_level(self, a_statement, is_last_statement_of_shape, namespaces_dict):
        tuples_line_indent = []
        st_property = BaseStatementSerializer.tune_token(a_statement.st_property, namespaces_dict)
        st_target_elements = []
        for a_type in a_statement.st_types:
            st_target_elements.append(self.str_of_target_element(target_element=a_type,
                                                                 st_property=a_statement.st_property,
                                                                 namespaces_dict=namespaces_dict))

        tuples_line_indent.append(FixedPropChoiceStatementSerializer._opening_tuple_line_of_choice())

        tuples_line_indent.append(FixedPropChoiceStatementSerializer.
                                  _statement_in_choice_no_cardinality(st_property,
                                                                      st_target_elements[0]))

        for a_type in st_target_elements[1:]:
            tuples_line_indent.append(FixedPropChoiceStatementSerializer._tuple_of_disjunction())
            tuples_line_indent.append(
                FixedPropChoiceStatementSerializer._statement_in_choice_no_cardinality(st_property,
                                                                                       a_type))

        tuples_line_indent.append(self._tuple_closing_choice(a_statement,
                                                             is_last_statement_of_shape))

        for a_comment in a_statement.comments:
            tuples_line_indent.append((a_comment, 4))
        a = 3 + 1

        return tuples_line_indent


    def _tuple_closing_choice(self, a_statement, is_last_statement_of_shape):
        str_res = ")" + SPACES_GAP_BETWEEN_TOKENS + \
                  BaseStatementSerializer.cardinality_representation(cardinality=a_statement.cardinality,
                                                                     statement=a_statement,
                                                                     out_of_comment=True) + \
                  BaseStatementSerializer.closure_of_statement(is_last_statement_of_shape)

        if a_statement.cardinality != KLEENE_CLOSURE and not self._disable_comments:
            str_res += BaseStatementSerializer.adequate_amount_of_final_spaces(str_res) + \
                       BaseStatementSerializer.probability_representation(a_statement.probability)
        return str_res, 1

    @staticmethod
    def _tuple_of_disjunction():
        return "|", 2

    @staticmethod
    def _opening_tuple_line_of_choice():
        return "(", 1

    @staticmethod
    def _statement_in_choice_no_cardinality(st_property, st_type):
        return (st_property + SPACES_GAP_BETWEEN_TOKENS + st_type, 1)

    @staticmethod
    def turn_statement_into_comment(statement, namespaces_dict):
        return statement.probability_representation() + \
               " with cardinality " + statement.cardinality_representation()



from shexer.utils.shapes import build_shapes_name_for_class_uri

def determine_original_target_nodes_if_needed(remove_empty_shapes, original_target_classes, original_shape_map):
    if not remove_empty_shapes:
        return None  # We dont need this structure if there are no shapes to remove.
    result = set()
    if original_target_classes is not None:
        for a_class in original_target_classes:
            result.add(build_shapes_name_for_class_uri(a_class))
    if original_shape_map is not None:
        for an_item in original_shape_map.yield_items():
            result.add(an_item.shape_label)
    return result



from shexer.io.line_reader.file_line_reader import FileLineReader
from shexer.io.line_reader.raw_string_line_reader import RawStringLineReader
from shexer.utils.obj_references import check_just_one_not_none

class BaseTriplesYielder(object):

    def __init__(self):
        pass

    def _decide_line_reader(self, raw_graph, source_file):
        check_just_one_not_none((source_file, "source_file"),
                                (raw_graph, "raw_graph"))
        if raw_graph is not None:
            return RawStringLineReader(raw_string=raw_graph)
        else:
            return FileLineReader(source_file=source_file)

    def yield_triples(self):
        raise NotImplementedError("Implement this method in derived classes")


from shexer.shaper import Shaper
from shexer.consts import TURTLE

a_graph = """
@prefix : <http://example.org/> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .

:Marie a :person ;
	:likes :Justine ;
	:likes :Antoine ;
	:likes :football ;
	:name "Marie" ;
	:age "30"^^xsd:int .

:Justine a :person ;
    :likes :Marie ;
    :likes :basketball ;
    :name "Justine" .

:Antoine a :person ;
	:likes :Marie ;
	:likes :Justine ;
	:name "Antoine" ;
	:age "32"^^xsd:int .
"""

shaper = Shaper(all_classes_mode=True,
                raw_graph=a_graph,
                input_format=TURTLE)
resutl = shaper.profile_graph(string_output=True)
print(resutl)